package jia;

import java.awt.Point;
import java.util.*;

import prop.PropertiesReader;

import env.WorldModel;

import arch.CowboyArch;
import arch.LocalWorldModel;
import jason.asSemantics.*;
import jason.asSyntax.*;
import jason.environment.grid.Location;

/**
 * Is used when the leader delegates a new target to the agent. If the agent
 * already has a target the revise_target should be used.
 * 
 */
public class delegate_target extends DefaultInternalAction {

	astar pathFinder;
	LocalWorldModel model;
	TransitionSystem ts;

	@Override
	public Object execute(TransitionSystem ts, Unifier un, Term[] args) throws Exception {
		model = ((CowboyArch) ts.getUserAgArch()).getModel();
		this.ts = ts;

		try {
			int agX = (int) ((NumberTerm) args[0]).solve();
			int agY = (int) ((NumberTerm) args[1]).solve();
			Atom currentAgentAtom = (Atom) args[2];
			String currentAgent = currentAgentAtom.getFunctor().equals("self") ? ts.getUserAgArch().getAgName() : currentAgentAtom.getFunctor();
			Point agent = new Point(agX, agY);

			int currentStep = (int) ((NumberTerm) args[3]).solve();

			SwitchFinder finder = new SwitchFinder(model);
			Point target;
			// Switch?
			target = finder.getSwitch(agent, currentAgent);

			// no switch should be triggered, find a target
			if (target == null)
				target = decideTarget(agent, currentAgent, currentStep);

			if (target != null) {
				model.setTargetAgentPos(CowboyArch.getAgId(currentAgent), target.x, target.y);
				return un.unifies(args[4], new NumberTermImpl(target.x)) && un.unifies(args[5], new NumberTermImpl(target.y));
			}

		} catch (Exception e) {
			e.printStackTrace();
			ts.getLogger().info("Error in delegate_target");
		}
		return false;
	}

	public Point decideTarget(Point agent, String agentName, int currentStep) {
		pathFinder = new astar();
		CowCluster delegatedCluster = null;

		Point target = new Point();
		Point corral = new Point(model.getCorralCenter().x, model.getCorralCenter().y);

		CowCluster[] clusters;
		int stepsBetweenRevisions = PropertiesReader.getInstance().getIntegerProperty("steps_between_revisions");
		if (Math.abs(currentStep - model.lastRevisionStep) > stepsBetweenRevisions || currentStep == 0) {
			model.lastRevisionStep = currentStep;
			int connectionRange = PropertiesReader.getInstance().getIntegerProperty("connection_range");
			int maxClusterSize = PropertiesReader.getInstance().getIntegerProperty("max_cluster_size");
			clusters = ConnectedComponents.getClusters(ts, cowArray(model), connectionRange, maxClusterSize);

			// calculate possible targets
			for (CowCluster c : clusters) {
//				ts.getLogger().info("clustercenter: " + c.x + "," + c.y + " childs: " + c.cows.size());

				calculateTargets(c, ts);

			}

			model.clusters = clusters;

			// finds the agents which have been delegated each target
			// returns the delegated cluster
			delegatedCluster = delegateTargets(agentName, clusters);
		}
		else {
			clusters = model.clusters;

			for (CowCluster c : clusters) {
				for (Target targ : c.targets) {
					if (targ.agent == CowboyArch.getAgId(agentName)) {
						delegatedCluster = c;
					}
				}
			}
		}

		// System.out.println(clusters.length);

		// If the agent did not have a cluster delegated - find a free target
		if (delegatedCluster == null) {
			// get cluster which is closest and has space for another agent
			double minDist = Double.MAX_VALUE;
			for (CowCluster c : clusters) {
				boolean clusterOccupied = true;
				for (Target t : c.targets) {
					if (t.agent == -1) // no agent delegated to this target
						clusterOccupied = false;
				}
				// if all targets are delegated, go to next cluster
				if (clusterOccupied)
					continue;

				double tmpDist = (agent.distance(c.x, c.y) / 2) + c.pathLength * 2;
				if (tmpDist < minDist) {
					minDist = tmpDist;
					delegatedCluster = c;
				}
			}
		}

		// if a free cluster is found
		if (delegatedCluster != null) {
			// Get a free target in the cluster
			// looks for the closest free target
			double minDist = Double.MAX_VALUE;
			Target chosenTarget = null;
			for (Target t : delegatedCluster.targets) {
				// If a target has been delegated to this agent already
				if (t.agent == CowboyArch.getAgId(agentName)) {
					chosenTarget = t;
					break;
				}
				// else select the target (which does not have an agent) closest
				// to the agent
				double tmpDist = t.target.distance(agent);
				if (tmpDist < minDist && t.agent == -1) {
					minDist = tmpDist;
					chosenTarget = t;
				}
			}
			target = chosenTarget.target;
		}
		else {
			// calculate a direction in which cows may be
			target = exploreMap(corral);
		}

		if (target.x < 0)
			target.x = 0;
		if (target.x >= model.getWidth())
			target.x = model.getWidth() - 1;
		if (target.y < 0)
			target.y = 0;
		if (target.y >= model.getHeight())
			target.y = model.getHeight() - 1;

		return target;
	}

	private CowCluster delegateTargets(String currentAgent, CowCluster[] clusters) {
		CowCluster delegatedCluster = null;
		// get currently delegated targets
		for (int i = 0; i < model.targets.length; i++) {
			int agentID = i;
			Location agentTarget = model.targets[i];
			if (agentTarget == null || agentTarget.x == -1)
				continue;

			Target closest = null;
			double minDist = Double.MAX_VALUE;
			CowCluster correspondingCluster = null;
			for (CowCluster c : clusters) {
				for (Target targ : c.targets) {
					double tmpDist = targ.target.distance(agentTarget.x, agentTarget.y);
					if (tmpDist < minDist && targ.agent == -1) {
						minDist = tmpDist;
						closest = targ;
						correspondingCluster = c;
					}
				}
			}

			// must be close to the cluster, otherwise it may result in weird mixups
			if (closest != null && minDist < 5) {
				closest.agent = agentID;
				model.setTargetAgentPos(agentID, closest.target.x, closest.target.y);

				// if the closest target is the agents own target, use this
				if (closest.agent == CowboyArch.getAgId(currentAgent))
					delegatedCluster = correspondingCluster;
			}
		}
		return delegatedCluster;
	}

	private void calculateTargets(CowCluster cluster, TransitionSystem ts) {
				
		Point clusterCenter = new Point(cluster.x, cluster.y);
		Location corralCenter= model.getCorralCenter();

		// the path these cows must move to get to the corral (using A*)
		ArrayList<Node> path = pathFinder.astar(ts, clusterCenter.x, clusterCenter.y, corralCenter.x, corralCenter.y, false);

		cluster.pathLength = path.size();
		
		
		Node pathNode = null;
		if (path.size() == 0)
			return;
		else if (path.size() < 6)
			pathNode = path.get(path.size()-1);
		else
			pathNode = path.get(5);
		
		
		
		Vec reverseVector = new Vec((clusterCenter.x - pathNode.x), (clusterCenter.y - pathNode.y));
		reverseVector = reverseVector.newMagnitude(2.0);
		
		int variableAgentDeviationDegree = PropertiesReader.getInstance().getIntegerProperty("agent_deviation");
		int variableDistanceToCows = PropertiesReader.getInstance().getIntegerProperty("distance_to_cows");
		
		ArrayList<Vec> vectorArray = calculateFormationVectors(cluster, reverseVector, variableAgentDeviationDegree);
		Point[] calculatedTargets = null;
		
		boolean minimumDistanceAcceptable = false;
		
		int loopCount = 0;
		while(minimumDistanceAcceptable == false) {
			
			calculatedTargets = new Point[vectorArray.size()];
		
			for (int i = 0; i < vectorArray.size(); i++) {
				
				int targetX = (int) ((double)clusterCenter.x + vectorArray.get(i).x);
				int targetY = (int) ((double)clusterCenter.y + vectorArray.get(i).y);
				
				Point targetPoint = new Point(targetX, targetY);
				
				double multiplyConstant = 1.1;
				while(distToNearestCow(targetPoint, cluster.cows) < variableDistanceToCows) {
					Point tempTarget = new Point();
					
					tempTarget.x = (int) ((double)clusterCenter.x + (vectorArray.get(i).x * multiplyConstant));
					tempTarget.y = (int) ((double)clusterCenter.y + (vectorArray.get(i).y * multiplyConstant));
					
					if (model.hasObject(WorldModel.OBSTACLE, tempTarget.x, tempTarget.y))
						break;
					else 
						targetPoint = new Point(tempTarget.x, tempTarget.y);
						
					multiplyConstant += 0.1;
				}
				
				// add target to calculatedTargets
				calculatedTargets[i] = targetPoint;
			}
			
			minimumDistanceAcceptable = true;
			for (int i = 0; i < calculatedTargets.length; i++) {
				for (int j = 0; j < calculatedTargets.length; j++) {
					if (i == j)
						continue;
					
					int distanceBetweenAgents = PropertiesReader.getInstance().getIntegerProperty("distance_between_agents");
					if (calculatedTargets[i].distance(calculatedTargets[j]) < distanceBetweenAgents && variableDistanceToCows < 5)
						minimumDistanceAcceptable = false;

				}
			}
			
//			if (minimumDistanceAcceptable) {
//				ts.getLogger().info("formation found at, clustersize: " + calculatedTargets.length + " degreeDeviation: " + variableAgentDeviationDegree + " distanceToCows: " + variableDistanceToCows);
//			}
			
			if (loopCount > 100) {	
//				ts.getLogger().info("loop count exeeded. degreeDeviation: " + variableAgentDeviationDegree + " distanceToCows: " + variableDistanceToCows);
				continue;
			}
			
			
			
			if (minimumDistanceAcceptable == false) {
				int degreeSpan = PropertiesReader.getInstance().getIntegerProperty("agents_degree_span");
				
				if (vectorArray.size() * variableAgentDeviationDegree < degreeSpan)
					vectorArray = calculateFormationVectors(cluster, reverseVector, (variableAgentDeviationDegree += 5));
				else {
					variableDistanceToCows++;
					variableAgentDeviationDegree = PropertiesReader.getInstance().getIntegerProperty("agent_deviation");
					vectorArray = calculateFormationVectors(cluster, reverseVector, variableAgentDeviationDegree);
				}
			}
			
			loopCount++;
		}

		// save target
		for (Point targetPoint : calculatedTargets)
			cluster.targets.add(new Target(targetPoint));

		
	}
	
	public ArrayList<Vec> calculateFormationVectors(CowCluster cluster, Vec reverseVector, int agentDeviationDegree) {
		
		ArrayList<Vec> vectorArray = new ArrayList<Vec>();
		
		if (cluster.cows.size() == 1) { // 1 agent
			vectorArray.add(reverseVector);
		} else if (cluster.cows.size() < 5) { // 2 agents
			vectorArray.add(Vec.rotate(reverseVector, -1*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, 1*agentDeviationDegree));
		} else if (cluster.cows.size() < 8) { // 3 agents
			vectorArray.add(reverseVector);
			vectorArray.add(Vec.rotate(reverseVector, -2*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, 2*agentDeviationDegree));
		} else if (cluster.cows.size() < 12) { // 4 agents
			vectorArray.add(Vec.rotate(reverseVector, -1*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, 1*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, -3*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, 3*agentDeviationDegree));
		} else if (cluster.cows.size() <= 15) { // 5 agents
			vectorArray.add(reverseVector);
			vectorArray.add(Vec.rotate(reverseVector, -2*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, 2*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, -4*agentDeviationDegree));
			vectorArray.add(Vec.rotate(reverseVector, 4*agentDeviationDegree));
		}
		
		return vectorArray;
	}
	
	public boolean directWayClear(Point clusterCenter, Node pathNode) {
		Vec vector = new Vec(pathNode.x - clusterCenter.x, pathNode.y - clusterCenter.y);
		vector = vector.newMagnitude(0.3);
		
		Point selectedPoint = new Point( (int) ((double)clusterCenter.x + vector.x), (int) ((double)clusterCenter.y + vector.y) );
		
		double multiplyConstant = 1.1;
		
		while (Point.distance(selectedPoint.x, selectedPoint.y, pathNode.x, pathNode.y) > 2) {
			selectedPoint = new Point( 	(int) ((double)clusterCenter.x + (vector.x * multiplyConstant)) , 
										(int) ((double)clusterCenter.y + (vector.y * multiplyConstant)) );
			
			if (model.hasObject(WorldModel.OBSTACLE, selectedPoint.x, selectedPoint.y)) {
				return false;
			}
			
			multiplyConstant += 0.1;
		}
		
		return true;
	
	}

	private double distToNearestCow(Point target, ArrayList<Vec> cows) {
		double minDist = Double.MAX_VALUE;
		for (Vec cow : cows) {
			double tmpDist = target.distance(cow.x, cow.y);
			if (tmpDist < minDist)
				minDist = tmpDist;
		}
		return minDist;
	}

	private int[][] cowArray(LocalWorldModel model) {
		int[][] cowsInMap = new int[model.getWidth()][model.getHeight()];
		for (int i = 0; i < model.getWidth(); i++) {
			for (int j = 0; j < model.getHeight(); j++) {
				if (model.hasObject(WorldModel.COW, i, j))
					cowsInMap[i][j] = 1;
			}
		}
		return cowsInMap;
	}

	private Point exploreMap(Point corral) {
		Random r = new Random(System.currentTimeMillis());
		Point target = null;
		int lowest = Integer.MAX_VALUE;
		for(int i = 0; i < 100; i++) {
			int x = r.nextInt(model.getWidth());
			int y = r.nextInt(model.getHeight());

			if (model.agentVisits[x][y] < lowest) {
				target = new Point(x, y);
				lowest = model.agentVisits[x][y];
			}
		}

		return target;
	}

}

class Target {
	public Target(Point target) {
		this.target = target;
	}
	
	public Target() {
		this.target = new Point();
	}
	
	Point target;
	int agent = -1;

	public String toString() {
		return "Agent" + agent + " -> " + target;
	}
}